﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;
using Ionic.Zip;
using ProcessJavaScript;
 
namespace BlanketyBlank
    
{
    
    public partial class Form1 : Form
    {
        /*
         * Called From:
         *      Form1()
         *      copyFromClipBtn_Click()
         *
         * This routine inserts the contents of the Clipboard into the inText text box and
         * and calls CreateBoxList() to build the box list from the boxes found in inText.
        */
        public void ProcessInputCode()
        {
            inText.Clear();
            outText.Clear();

            /*
             * Any code editor selected code has been copied to clipboard:
             */
            inText.Text = (Clipboard.ContainsText(TextDataFormat.Text) ? Clipboard.GetText(TextDataFormat.Text) : "");

            // Now create the box list from the selected text.
            CreateBoxList();

            // Setting focus to the code input area.
            inText.Select();
        }

        /*
         * Called From:
         *      ProcessInputCode()
         *      codeBoxListBtn_Click()
         *      
         * Clear and then populate the codeBoxList with all boxes
         * found in the inText RichTextBox. Also clear the 'child'
         * list codeExclBoxList.
        */
        public void CreateBoxList()
        {
            List<string> inTextBoxList = new List<string>();
            string inputStr;
            MatchCollection matches;

            codeBoxList.Items.Clear();
            codeExclBoxList.Items.Clear();

            inputStr = inText.Text;

            //MessageBox.Show(inputStr);
            

            if (inputStr.Length > 0)
            {
                /*
                 * Now get all boxes in inputStr and insert into inTextBoxList avoiding
                 * duplicate boxes. We use Regex.Matches because we expect many boxes as
                 * .Matches returns a MatchCollection.
                 */
                matches = Regex.Matches(inputStr, @"(box\d+_\d+|box\d+)");
                
                // Iterate over the collection of matches:
                foreach (Match match in matches)
                {
                    // MessageBox.Show(String.Format("match.Value: {0}", match.Value));
                    // Insert the box if it is not in the list already:
                    if (!inTextBoxList.Contains(match.Value))
                    {
                        inTextBoxList.Add(match.Value);
                    }
                }

                // First check if we have anything in our local box list:
                if (inTextBoxList.Count > 0)
                {
                    // Now sort the list:
                    inTextBoxList.Sort(new BoxCompare());

                    // Now populate codeBoxList from inTextBoxList and select the first item:
                    foreach (string item in inTextBoxList)
                    {
                        codeBoxList.Items.Add(item);
                    }
                    codeBoxList.SetSelected(0, true);
                }
            }
        }

        // Clears inText, outText, codeBoxList, and codeExclBoxList.
        public void ClearInputOutputCodeAndLists()
        {
            inText.Clear();
            outText.Clear();

            codeBoxList.Items.Clear();
            codeExclBoxList.Items.Clear();
        }

        /*
         * Called From:
         *      Form1()
         *      javascriptToolStripMenuItem_Click()
         * 
         * This reads the documents .ZIP file to build the document's sorted box list.
         * The Box List label will also display the file name minus the extension.
         * 
         * Callers have ensured zipFile is not null or empty.
        */
        public void ProcessDocFile(string zipFile)
        {
            string xmlFile = "";

            try
            {
                curDir = Path.GetDirectoryName(zipFile);

                xmlFile = UnzipXmlFile(zipFile);

                if (!String.IsNullOrEmpty(xmlFile))
                {
                    docBoxListLbl.Text = "Box List " + Path.GetFileNameWithoutExtension(zipFile).ToUpper();

                    CreateDocBoxList(xmlFile);

                    File.Delete(xmlFile);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        /*
         * This extracts the "DOC.XML" file from the ZIP file identified by the
         * argument zipFileName.
         *
         * Called From:
         *  ProcessDocFile()
         *
         * Inputs:
         *  zipFileName: The full path and file name of the zip file.
         *
         * Returns the full path and file name of the extracted "DOC.XML" file or null
         *  if the exraction process did not work out well.
        */
        public string UnzipXmlFile(string zipFileName)
        {
            string docXmlFile = null;
            string tmpPath;
            // Got this password from Joe A.
            string password = "5e$ACre7ru";


            try
            {
                // We first need to get the temp path where we will store the extracted XML file, the
                // caller will delete the file after reading it for checkbox information.
                tmpPath = Path.GetTempPath();

                if (!String.IsNullOrEmpty(tmpPath))
                {
                    //MessageBox.Show(string.Format("UnzipXmlFile - tmpPath: {0}", tmpPath));

                    using (ZipFile zip = ZipFile.Read(zipFileName))
                    {
                        foreach (ZipEntry e in zip)
                        {
                            if (e.FileName.ToLower() == "doc.xml")
                            {
                                if (e.UsesEncryption)
                                {
                                    e.Password = password;
                                }
                                e.Extract(tmpPath, ExtractExistingFileAction.OverwriteSilently);

                                // Set the path and file name of the extracted XML file to be returned to caller.
                                docXmlFile = tmpPath + "doc.xml";
                            }
                        }
                    }
                }
            }
            catch (System.Security.SecurityException ex)
            {
                MessageBox.Show(ex.ToString(), "UnzipXmlFile(): Could not get path to Temp folder.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return docXmlFile;
        }

        /*
         * Called From:
         *      ProcessDocFile()
         *      
         * Clear and then populate the docBoxList with all boxes
         * found in the document's .Zip file.
         * 
         * Also clear the 'child' list docInclBoxList.
         * 
        */

        //################################################################
        //
        // The method below reads the form XML and creates a box list.
        // The reason for the box list is unclear as a query of the 
        // calc'ers revealed no one is using it but it's here if anyone
        // figures out a use for it.
        //
        //################################################################

        public void CreateDocBoxList(string xmlFile)
        {
            List<string> xmlBoxList = new List<string>();
            string line;
            Match match;


            docBoxList.Items.Clear();
            docInclBoxList.Items.Clear();

            try
            {
                using (StreamReader sr = new StreamReader(xmlFile))
                {
                    while ((line = sr.ReadLine()) != null)
                    {
                        // Remove whitespace:
                        line = Regex.Replace(line, @"\s+", "");
                        //MessageBox.Show(line);

                        // We loop through each line of the XML file until a box
                        // matching the regex below is detected.
                        // The box is then added to the xmlBoxList.Add(match.Groups[1].Value);
                        // Once the end of the xml file is reached, if there were any
                        // box matches the xmlBoxList is passed to the BoxCompare()
                        // method in CustomClasses.cs to sort it into ascending order.

                        // Got a match? "my socks and your mother's meatloaf!" [T. Frost R.I.P.]
                        match = Regex.Match(line, @"ref=""(box\d+_\d+|box\d+)""");

                        //MessageBox.Show("Display match.ToString()" + match.ToString());


                        if ((match.Success) && (match.Groups[1].Value != "box1_65535"))
                        {
                            //MessageBox.Show("Match Success");
                            xmlBoxList.Add(match.Groups[1].Value);
                            //MessageBox.Show("xmlBoxList.Add");

                        }
                    }
                }

                //The sort and BoxCompare places the boxes
                // in ascending order. 

                if (xmlBoxList.Count > 0)
                {
                    //#######################################
                    //
                    // This is the call to the BoxCompare
                    // function we mentioned above.
                    //
                    //#######################################

                    //BoxCompare Sort
                    //MessageBox.Show("xmlBoxList.sort");
                    xmlBoxList.Sort(new BoxCompare());

                    foreach (string box in xmlBoxList)
                    {
                        docBoxList.Items.Add(box);
                    }
                    docBoxList.SetSelected(0, true);
                }
            }
            catch (IOException ex)
            {
                MessageBox.Show(ex.ToString(), "CreateDocBoxList(): File could not be read.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                //MessageBox.Show(ex.ToString());
            }
        }

        /*
         * This uses the OpenFileDialog class in allowing the user to select
         * a zip file which will be used in retrieving document boxes.
         *
         * Called From:
         *   javascriptToolStripMenuItem_Click
         *
         * Inputs:
         *   dir: The initial file path from which a file may be selected.
         *
         * Returns the full path and file name of the selected file or null if no
         * file was selected.
        */
        public string GetFileName(string dir)
        {
            string fileName = null;
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.InitialDirectory = dir;

            ofd.Filter = "ZIP file|*.zip";

            System.Windows.Forms.DialogResult dr = ofd.ShowDialog();

            if (dr == DialogResult.OK)
            {
                // This holds the path and the file name with extension.
                fileName = ofd.FileName;
            }

            return fileName;
        }

        /*
         * This will reverse 'Blankety Blank'ed code to the original form.
         * Any code that can not be reversed will be regurgitated as received.
         * All of the work is actually done by a Javascript script file.
         * 
         * Called From:
         *   revBlnkBtn_Click()
        */
        public void ReverseBlank()
        {
            JavaScriptExpression jsExpr;

            outText.Clear();

            if (inText.Text.Trim().Length == 0)
            {
                // See ya!
                return;
            }

            /*
             * The Javascript script file does all of the work.
             */
            jsExpr = new JavaScriptExpression("RevBlank", revBlankTmp);

            if (!jsExpr.CheckArgs("RevBlank"))
            {
                // See ya!
                if (fromEditor)
                {
                    Application.Exit();
                }

                return;
            }

            try
            {
                File.WriteAllText(revBlankCode, inText.Text);

                if (!jsExpr.DoRevBlank(revBlankCode))
                {
                    // See ya!
                    if (fromEditor)
                    {
                        Application.Exit();
                    }

                    return;
                }

                outText.Text = File.ReadAllText(revBlankCode);

                // Setting focus to the code output area.
                outText.Select();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return;
            }
        }

       

        /*
         * This performs single or multiple blanking. Actually, the preparations
         * for blanking are done here, actual blanking is done by
         * DoTheBlankThing().
         *
         * Called From:
         *   singleBlnkBtn_Click()
         *   multiBlnkBtn_Click()
         *   custBlnkBtn_Click()
         *
         * Inputs:
         *   mode: Indicates either "single", "multiple", or "custom" blanking is
         *   to be performed.
        */
        public void PrepareToBlank(string mode)
        {
            JavaScriptExpression jsExpr;

            List<string> blankList = new List<string>();
            List<CodeBlock> codeList = new List<CodeBlock>();

            CodeBlock codeBlock = null;
            CodeBlock tmpCodeBlock = null;

            string tmpStr = "";
            string printBox = "";
            string box = "";
            string line = "";
            string errMsgCaption = "PrepareToBlank(): Ambiguity error.";
            string errMsg = "";

            string[] linesArray;
            string[] tmpArray;

            int groupCnt = 1;
            int groupNum = 0;
            int colon = -1;
            int startIndex = -1;
            int startVarIndex = -1;
            int stopIndex = -1;
            int stopVarIndex = -1;
            int boxesIndex = 0;
            int chkVarLastIndex = -1;

            bool chkVarAssign = false;
            bool chkVarIndex = false;
            bool setIndex = false;

            Dictionary<int, int> vars = new Dictionary<int, int>();
            Dictionary<int, int> boxes = new Dictionary<int, int>();

            Match match;

            outText.Clear();

            //MessageBox.Show("Into Multi-blanking");

            if (inText.Text.Trim().Length == 0)
            {
                // See ya!
                return;
            }

            /*
             * This will strip out the hasOverrides 'if' expressions and will
             * also gather the starting and stopping index values of variable
             * and box assignments using the CompleteExpression() routine which
             * is housed in the JS file 'preBlankTmp'.
             * 
             * The indices are the array indices of the code array [the code to
             * be blanked] after the override 'if's have been removed as well
             * as blank lines.
             */
            jsExpr = new JavaScriptExpression("PreBlank", preBlankTmp);

            if (!jsExpr.CheckArgs("PreBlank"))
            {
                // See ya!
                if (fromEditor)
                {
                    Application.Exit();
                }

                return;
            }

            try
            {
                File.WriteAllText(preBlankCode, inText.Text);

                if (!jsExpr.DoPreBlank(preBlankCode))
                {
                    // See ya!
                    if (fromEditor)
                    {
                        Application.Exit();
                    }

                    return;
                }

                /*
                 * We must not delete any lines else we mess up our variable
                 * and box assignment dictionaries.
                 * 
                 * The first element in this array MAY be a string detailing
                 * all variable assignments encountered in the code, the starting
                 * array index number followed by a colon then the stopping index
                 * number of the assignment. Multiple variable assignments are
                 * separated by the pipe character '|'.
                 * 
                 * If no variable assignments were found, then the first element
                 * will be for all box assignments with the same structure used.
                 * Vars will have the 'VARS::' leading chars, boxes will have
                 * 'BOXES::'. These index values will be adjusted for the presence
                 * of these two strings, i.e. we keep the starings in the array.
                 */
                linesArray = File.ReadAllText(preBlankCode).Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

                //MessageBox.Show("Into linesArray.Length > 0");

                if (linesArray.Length > 0)
                    
                {
                    

                    if (linesArray[0].StartsWith("VARS::"))
                    {
                        MessageBox.Show("linesArray[0]");
                        boxesIndex += 1;
                        tmpArray = linesArray[0].Substring(6).Split('|');

                        

                        foreach (string var in tmpArray)
                        {
                            colon = var.IndexOf(":");

                            if (colon != -1)
                            {
                                int.TryParse(var.Substring(0, colon), out startIndex);
                                int.TryParse(var.Substring(colon + 1), out stopIndex);
                                vars.Add(startIndex + 2, stopIndex + 2);
                            }
                            else
                            {
                                // Error message and return.
                                errMsg = "  Variable index missing ':'\n\n  Aborting blanking operation.";
                                ProcessAmbiguityError(errMsg, errMsgCaption);
                                return;
                            }
                        }
                        
                    }

                    // The JS file that does the work errors out if there are no boxes.
                    if (linesArray[boxesIndex].StartsWith("BOXES::"))
                    {
                        int offset = (boxesIndex > 0 ? 2 : 1);

                        tmpArray = linesArray[boxesIndex].Substring(7).Split('|');

                        foreach (string assignBox in tmpArray)
                        {
                            colon = assignBox.IndexOf(":");

                            if (colon != -1)
                            {
                                int.TryParse(assignBox.Substring(0, colon), out startIndex);
                                int.TryParse(assignBox.Substring(colon + 1), out stopIndex);
                                boxes.Add(startIndex + offset, stopIndex + offset);
                            }
                            else
                            {
                                // Error message and return.
                                errMsg = "  Box index missing ':'\n\n  Aborting blanking operation.";
                                ProcessAmbiguityError(errMsg, errMsgCaption);
                                return;
                            }
                        }
                    }
                }
                else
                {
                    // Error message and return.
                    errMsg = "  Empty linesArray after stripping Overrides.\n\n  Aborting blanking operation.";
                    ProcessAmbiguityError(errMsg, errMsgCaption);
                    return;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return;
            }

            tmpStr = Regex.Replace(groupSizeEntry.Text, @"\s+", "");

            match = Regex.Match(tmpStr, @"^([+]?\d+)(\.0*)?$");
            if (match.Success)
            {
                //MessageBox.Show(String.Format("match.Groups[1].Value: {0}", match.Groups[1].Value));
                int.TryParse(match.Groups[1].Value, out groupNum);
            }

            for (int i = boxesIndex + 1; i < linesArray.Length; i++)
            {
                line = linesArray[i];

                /*
                 * Need to check for a var assignment that spans more than
                 * one line.
                 */
                if (chkVarAssign)
                {
                    if (stopVarIndex == i)
                    {
                        chkVarAssign = false;
                    }

                    codeBlock.codeLine.Add(line);

                    if (codeBlock.blockType == blockTypes.calc)
                    {
                        codeBlock.codeIndex.Add(i);
                    }

                    continue;
                }

                /*
                 * Just catching a goofy 'unfinished code block error message' as an
                 * empty line (after trimming) will default into the current codeBlock.
                 * 
                 * SHOULD BE ABLE TO REMOVE THIS...CHECK FIRST WHEN READY............
                 */
                //if (line.Length == 0) continue;

                /*
                 * Check if we have a print line - this flags the end of a codeBlock as it
                 * would be the last thing you do for a codeBlock.
                 */
                match = Regex.Match(line.ToLower(), @"sfforms.storagemanager.setvalue\s*\(\s*forminstanceid,\s*'(box\d+_\d+|box\d+)',\s*(box\d+_\d+|box\d+)\s*\);");

                if (match.Success)
                {
                    printBox = match.Groups[1].Value;

                    /*
                     * First we do a set of error checks:
                     * 
                     * 1) We must have an instantiated codeBlock by now and...
                     *      1a) The codeBlock must have an assigned calcBox;
                     *      1b) The printBox must equal the codeBlock.calcBox;
                     */
                    errMsg = "";

                    if (codeBlock == null)
                    {
                        errMsg = string.Format("  Print line box: {0}\n  does not have an associated CodeBlock\n\n  Aborting blanking operation.", printBox);
                    }
                    else if (codeBlock.calcBox == null)
                    {
                        errMsg = string.Format("  Print line box: {0}\n  does not have an associated CodeBlock.calcBox,\n  it is NULL.\n\n  Aborting blanking operation.", printBox);
                    }
                    else if (codeBlock.calcBox != printBox)
                    {
                        errMsg = string.Format("  Print line box: {0}\n  does not equal the CodeBlock.calcBox: {1}.\n\n  Aborting blanking operation.", printBox, codeBlock.calcBox);
                    }

                    if (errMsg.Length > 0)
                    {
                        ProcessAmbiguityError(errMsg, errMsgCaption);
                        return;
                    }
                    else
                    {
                        /*
                         * Now check for contiguous variable assignments and comment
                         * lines on the 'outside' edge of the codeBlock, i.e. we can
                         * pull these out of the present codeBlock and make a new
                         * nonCalc codeBlock for these which will follow the edited
                         * .calc codeBlock. Note that we are talking here about
                         * trailing edge not leading edge, they were grabbed as we
                         * encountered them.
                         * 
                         * Doing this will avoid wrapping any var assignments
                         * [here we are referring to the 'outside' ones] inside the
                         * override 'if', or in the case of one or more comments,
                         * could reduce the codeBlock's code to the simpler ternary
                         * version of doBlank.
                         */
                        chkVarIndex = false;
                        setIndex = false;
                        startVarIndex = -1;
                        chkVarLastIndex = -1;

                        for (int j = 0; j < codeBlock.codeLine.Count; j++)
                        {
                            if (chkVarIndex)
                            {
                                if (codeBlock.codeIndex[j] == chkVarLastIndex)
                                {
                                    chkVarIndex = false;
                                }
                            }
                            else if (Regex.IsMatch(codeBlock.codeLine[j], @"^\/\/"))
                            {
                                setIndex = true;
                            }
                            else if (vars.ContainsKey(codeBlock.codeIndex[j]))
                            {
                                setIndex = true;

                                if (codeBlock.codeIndex[j] != vars[codeBlock.codeIndex[j]])
                                {
                                    chkVarLastIndex = vars[codeBlock.codeIndex[j]];
                                    chkVarIndex = true;
                                }
                            }
                            else
                            {
                                startVarIndex = -1;
                            }

                            if (setIndex)
                            {
                                if (startVarIndex == -1)
                                {
                                    startVarIndex = j;
                                }

                                setIndex = false;
                            }
                        }

                        /*
                         * Now check for 'outside' edge comment/var assignment code lines,
                         * if any move them out of the Calc codeBlock into a new nonCalc
                         * temp CodeBlock. Also check for any var assignments still left
                         * in the .calc codeBlock and set .hasVar accordingly,
                         * DoTheBlankThing() will process 'internal' vars.
                         */
                        if (startVarIndex > -1)
                        {
                            int count = codeBlock.codeLine.Count - startVarIndex;

                            tmpCodeBlock = new CodeBlock();
                            tmpCodeBlock.blockType = blockTypes.nonCalc;
                            tmpCodeBlock.codeLine = codeBlock.codeLine.GetRange(startVarIndex, count);

                            codeBlock.codeLine.RemoveRange(startVarIndex, count);
                            codeBlock.codeIndex.RemoveRange(startVarIndex, count);
                            codeBlock.hasVar = false;

                            for (int j = 0; j < codeBlock.codeLine.Count; j++)
                            {
                                if (vars.ContainsKey(codeBlock.codeIndex[j]))
                                {
                                    codeBlock.hasVar = true;
                                    break;
                                }
                            }
                        }

                        codeList.Add(codeBlock);

                        if (tmpCodeBlock != null)
                        {
                            codeList.Add(tmpCodeBlock);
                            tmpCodeBlock = null;
                        }

                        if (!blankList.Contains(printBox))
                        {
                            blankList.Add(printBox);
                        }

                        if (mode == "multiple")
                        {
                            /*
                             * We are multi-blanking, we need to determine if it is time
                             * to do it now, as print lines are significant.
                             */
                            if ((groupNum == 0) || (groupNum == groupCnt))
                            {
                                // Might not sort this...keep as an on-discovered basis...
                                // print lines will [should] match.

                                //blankList.Sort(new BoxCompare());

                                DoTheBlankThing(mode, ref blankList, ref codeList, ref boxes);

                                // Init for next group:
                                groupCnt = 1;
                                blankList.Clear();
                                codeList.Clear();
                            }
                            else
                            {
                                groupCnt += 1;
                            }
                        }

                        // Setup for next codeBlock:
                        codeBlock = null;
                        printBox = "";
                    }
                } // Print line stop.
                else
                {
                    // Looking for a comment line.
                    if (Regex.IsMatch(line, @"^\/\/"))
                    {
                        if (codeBlock == null)
                        {
                            codeBlock = new CodeBlock();
                            codeBlock.blockType = blockTypes.nonCalc;
                        }
                        else if (codeBlock.blockType == blockTypes.calc)
                        {
                            codeBlock.codeIndex.Add(i);
                        }

                        codeBlock.codeLine.Add(line);
                    }
                    else
                    {
                        /*
                         * Looking for a variable assignment line. We also consider
                         * that a variable assignment may span more than 1 line.
                         */
                        if (vars.ContainsKey(i))
                        {
                            if (codeBlock == null)
                            {
                                codeBlock = new CodeBlock();
                                codeBlock.blockType = blockTypes.nonCalc;
                            }
                            else if (codeBlock.blockType == blockTypes.calc)
                            {
                                codeBlock.hasVar = true;
                                codeBlock.codeIndex.Add(i);
                            }

                            stopVarIndex = vars[i];

                            if (stopVarIndex != i)
                            {
                                chkVarAssign = true;
                            }

                            codeBlock.codeLine.Add(line);
                        }
                        else
                        {
                            if (codeBlock == null)
                            {
                                codeBlock = new CodeBlock();
                                codeBlock.blockType = blockTypes.calc;
                            }
                            else if (codeBlock.blockType == blockTypes.nonCalc)
                            {
                                // Finish a pre-existing nonCalc CodeBlock.
                                codeList.Add(codeBlock);

                                codeBlock = new CodeBlock();
                                codeBlock.blockType = blockTypes.calc;
                            }

                            /*
                             * Check for a box assignment:
                             */
                            match = Regex.Match(line.ToLower(), @"^(box\d+_\d+|box\d+)\s+=\s+");

                            if (match.Success)
                            {
                                box = match.Groups[1].Value;

                                if (codeBlock.calcBox != null)
                                {
                                    if (box != codeBlock.calcBox)
                                    {
                                        errMsg = string.Format("  Calculated box: {0}\n  does not equal the CodeBlock box: {1}\n\n  Aborting blanking operation.", box, codeBlock.calcBox);

                                        ProcessAmbiguityError(errMsg, errMsgCaption);
                                        return;
                                    }
                                }
                                else
                                {
                                    codeBlock.calcBox = box;
                                }
                            }

                            codeBlock.codeLine.Add(line);
                            codeBlock.codeIndex.Add(i);
                        }
                    }
                }
            }

            // Debug - start:
            //foreach (CodeBlock block in codeList)
            //{
            //    outText.AppendText(string.Format("*****\nBlock Type: {0}\nCalc Box: {1}\nCode Line(s):\n", block.blockType, block.calcBox));

            //    foreach (string codeLine in block.codeLine)
            //    {
            //        outText.AppendText(string.Format("\t{0}\n", codeLine));
            //    }

            //    outText.AppendText("*****\n\n");
            //}
            //return;
            // Debug - stop:

            if (codeBlock != null)
            {
                /*
                 * It may be possible that we have a dangling 'nonCalc' codeBlock in
                 * which case we will add it to the codeList. A dangling 'calc' or
                 * 'unknown' codeBlock is an error, the user has not fully selected
                 * the code to blank.
                 */
                if (codeBlock.blockType == blockTypes.nonCalc)
                {
                    codeList.Add(codeBlock);
                }
                else
                {
                    errMsg = "  An unfinished codeBlock has been detected.\n\n  Make sure you have fully selected the code to blank\n  then try again.\n\n  Aborting blanking operation.";

                    ProcessAmbiguityError(errMsg, errMsgCaption);
                    return;
                }
            }

            /*
             * Check for any blanking that needs to be done, if we are single
             * blanking we will need to do it now, and there may be a blanking
             * group of codeBlocks that are not at the Group Size amount.
             * 
             * NOTE: Do not DoTheBlankThing() for a single .nonCalc codeBlock,
             *          we will do that right here.
             */
            if ((blankList.Count + codeList.Count) > 0)
            {
                if ((codeList.Count == 1) && (codeList[0].blockType == blockTypes.nonCalc))
                {
                    outText.AppendText((String.Join("\n", codeList[0].codeLine)) + "\n");
                }
                else
                {
                    DoTheBlankThing(mode, ref blankList, ref codeList, ref boxes);
                }
            }

            // Setting focus to the code output area.
            outText.Select();
        }

        public void ProcessAmbiguityError(string errorMsg, string caption)
        {
            MessageBox.Show(errorMsg, caption, MessageBoxButtons.OK, MessageBoxIcon.Error);

            if (fromEditor)
            {
                Application.Exit();
            }
        }

        /*
         * This performs the actual generation of blanking code. We generate the print
         * lines from blankList.
         *
         * Called From:
         *   PrepareToBlank()
         *
         * Inputs:
         *  mode:       Indicates either "single", "multiple", or "custom" blanking is
         *                  to be performed.
         *  blankList:  The list of boxes we found in the print lines.
         *  codeList:   The code lines that were not print lines, override lines, or '}' lines
         *                  that ended an override line.
         *  boxes:      The dictionary of box assignment starting and stopping indices
         *              [in PrepareToBlank()'s linesArray].
        */
        public void DoTheBlankThing(string mode, ref List<string> blankList, ref List<CodeBlock> codeList, ref Dictionary<int, int> boxes)
        {
            string boxTriggerStr = "";
            string outputStr = "";
            string errMsg = "";
            string errMsgCaption = "DoTheBlankThing(): Ambiguity error.";
            List<string> boxTriggerList = new List<string>();
            Match assignBox;
            MatchCollection matches;

            if (mode == "custom")
            {
                if (custBlankEntry.Text.Trim().Length > 0)
                {
                    boxTriggerStr = custBlankEntry.Text.Trim();

                    if (!Regex.Match(boxTriggerStr, @"^\(.+\)$").Success)
                    {
                        boxTriggerStr = "(" + boxTriggerStr + ")";
                    }
                }
                else
                {
                    MessageBox.Show("  Can not perform custom blanking:\n\n  'Custom BLANKETY BLANK Condition' text box is empty.", "DoTheBlankThing(): Missing custom blanking text.", MessageBoxButtons.OK, MessageBoxIcon.Error);

                    return;
                }
            }
            else
            {
                /*
                 * Not doing custom blanking, so build the boxStr for the boolean expression
                 * from the JavaScript code in the codeBlocks.
                 */
                foreach (CodeBlock codeBlock in codeList)
                {
                    foreach (string codeLine in codeBlock.codeLine)
                    {
                        // Do not use comment boxes as trigger boxes.
                        if (!Regex.IsMatch(codeLine, @"^\s*\/\/"))
                        {
                            matches = Regex.Matches(codeLine, @"(box\d+_\d+|box\d+)");

                            foreach (Match match in matches)
                            {
                                // MessageBox.Show(String.Format("match.Value: {0}", match.Value));
                                if (!(blankList.Contains(match.Value) || codeExclBoxList.Items.Contains(match.Value) || boxTriggerList.Contains(match.Value)))
                                {
                                    boxTriggerList.Add(match.Value);
                                }
                            }
                        }
                    }
                }

                if (mode != "multiple")
                {
                    /*
                     * Not multi-blanking, so adding any inclusion boxes.
                     * We have already made sure that the Exclusion and
                     * Inclusion box lists do not intersect. See the event
                     * handlers for the 'Exclusion Boxes' and 'INclusion Boxes'
                     * buttons for details.
                     */
                    foreach (string box in docInclBoxList.Items)
                    {
                        if (blankList.Contains(box))
                        {
                            MessageBox.Show(string.Format("  {0} cannot be a trigger box,\n  This is a calculated box.", box), "DoTheBlankThing(): Box is calculated.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                        else if (!boxTriggerList.Contains(box))
                        {
                            boxTriggerList.Add(box);
                        }
                    }
                }

                if (boxTriggerList.Count > 0)
                {
                    boxTriggerList.Sort(new BoxCompare());

                    boxTriggerStr = String.Join(" + ", boxTriggerList);
                }
                else
                {
                    MessageBox.Show("  Can not perform blanking:\n\n  No trigger boxes for blanking have been detected.", "DoTheBlankThing(): No trigger boxes detected.", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    if (fromEditor) Application.Exit();

                    return;
                }
            } // if (mode == "custom") {..} else {...

            // Now output the blanked code.
            if (mode == "custom")
            {
                outputStr = string.Format("// Blank START **********\ndoBlank = {0};\n\n", boxTriggerStr);
            }
            else
            {
                outputStr = string.Format("// Blank START **********\ndoBlank = (({0}).toString().trim().length == 0);\n\n", boxTriggerStr);
            }

            foreach (CodeBlock codeBlock in codeList)
            {
                if (codeBlock.blockType == blockTypes.nonCalc)
                {
                    foreach (string codeLine in codeBlock.codeLine)
                    {
                        outputStr += (string.Format("{0}\n", codeLine));
                    }
                }
                else if (codeBlock.blockType == blockTypes.calc)
                {
                    /*
                     * Need to check for a multi-line codeBlock with 'internal'
                     * variable assignment(s), if we have one we will wrap each
                     * box assignment in an override 'if', same approach as used
                     * in reverse blanking. This entails a different structure,
                     * and will be pretty rare, but at least no vars will be
                     * 'locked out' due to an override.
                     */
                    if ((codeBlock.codeLine.Count > 1) && (codeBlock.hasVar))
                    {
                        bool chkBoxIndex = false;
                        int chkBoxLastIndex = -1;
                        
                        outputStr += ("if (doBlank) {\n");
                        outputStr += (string.Format("{0} = '';\n", codeBlock.calcBox));
                        outputStr += "}\n";
                        outputStr += "else {\n";

                        for (int i = 0; i < codeBlock.codeLine.Count; i++)
                        {
                            if (chkBoxIndex)
                            {
                                if (codeBlock.codeIndex[i] == chkBoxLastIndex)
                                {
                                    // Got the last box assignment line. End the override too.
                                    chkBoxIndex = false;
                                    outputStr += (string.Format("{0}\n", codeBlock.codeLine[i]) + "}\n");
                                }
                                else
                                {
                                    outputStr += (string.Format("{0}\n", codeBlock.codeLine[i]));
                                }
                            }
                            else if (boxes.ContainsKey(codeBlock.codeIndex[i]))
                            {
                                // Wrap the box assignment in the override 'if'.
                                outputStr += (string.Format("if (!SFForms.StorageManager.hasOverrides(formInstanceId, \"{0}\")) ", codeBlock.calcBox) + "{\n");

                                outputStr += (string.Format("{0}\n", codeBlock.codeLine[i]));

                                if (codeBlock.codeIndex[i] != boxes[codeBlock.codeIndex[i]])
                                {
                                    chkBoxLastIndex = boxes[codeBlock.codeIndex[i]];
                                    chkBoxIndex = true;
                                }
                                else
                                {
                                    // All on one line, close the override 'if'.
                                    outputStr += "}\n";
                                }
                            }
                            else
                            {
                                outputStr += (string.Format("{0}\n", codeBlock.codeLine[i]));
                            }
                        }

                        // Close the doBlank 'if':
                        outputStr += "}\n";
                    }
                    else
                    {
                        outputStr += (string.Format("if (!SFForms.StorageManager.hasOverrides(formInstanceId, \"{0}\")) ", codeBlock.calcBox) + "{\n");

                        if (codeBlock.codeLine.Count > 1)
                        {
                            outputStr += ("if (doBlank) {\n");
                            outputStr += (string.Format("{0} = '';\n", codeBlock.calcBox));
                            outputStr += "}\n";
                            outputStr += "else {\n";

                            foreach (string codeLine in codeBlock.codeLine)
                            {
                                outputStr += (string.Format("{0}\n", codeLine));
                            }

                            outputStr += "}\n";
                        }
                        else
                        {
                            // Simple ternary doBlank version:
                            assignBox = Regex.Match(codeBlock.codeLine[0], @"^\s*(box\d+_\d+|box\d+)\s+=\s+(.+)$");

                            if (assignBox.Success)
                            {
                                // Take off the ending ';' from codeBlock.codeLine[0] - if present.
                                outputStr += string.Format("{0} = (doBlank ? '' : {1});\n", assignBox.Groups[1].Value, assignBox.Groups[2].Value.TrimEnd(';'));
                            }
                            else
                            {
                                // Should never happen......
                                errMsg = string.Format("  Expecting a box assignment but received:\n\n  {0}\n\n  Aborting blanking operation.", codeBlock.codeLine[0]);

                                ProcessAmbiguityError(errMsg, errMsgCaption);

                                if (fromEditor) Application.Exit();

                                return;
                            }
                        }

                        // Closing brace for if .has Overrides.
                        outputStr += "}\n";
                    }
                }
                else
                {
                    // Error out - unknown CodeBlock.
                    MessageBox.Show("  Can not perform blanking:\n\n  Unknown CodeBlock detected.", "DoTheBlankThing(): Unknown CodeBlock detected.", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    if (fromEditor) Application.Exit();

                    return;
                }

                // Add a blank line for ease of viewing.
                outputStr += "\n";
            }

            // Now the print lines - print lines feed this list:
            foreach (string blankBox in blankList)
            {
                outputStr += string.Format("SFForms.StorageManager.setValue(formInstanceId, '{0}', {1});\n", blankBox, blankBox);
            }

            /*
             * Auto-Format on paste in VSC does not indent this line
             * for some reason [WHEN IT IS THE VERY LAST LINE], so
             * forcing in 3 tabs.
             */
            outputStr += "\t\t\t// Blank STOP **********\n\n";

            outText.AppendText(outputStr);
        }
    }
}